/*------------------------------------------------------------------------
Apo to Biblio:
I.Vlahavas, P.Kefalas, N.Basileiadis, F.Kokkoras, I.Sakellariou
Texnhth Nohmosunh (B' Ekdosh), Ekdoseis Gartaganh, Thessaloniki, 2005
http://aibook.csd.auth.gr
------------------------------------------------------------------------*/
/*--------------------------------------------------------------------
Responding to a subset of KQML messages (see parsekqml.pl)

If the message is not syntactically correct then nothing happens.
Otherwise the parser returs the term:
message(performative(P), sender(S), receiver(R), replyto(RT),
	replywith(RW), language(L), content(C))
We then need to find the performative and if it is one that requires
action, we should produce a reply message. The reply message is saved
in a file in the dircetory of the receiver.
--------------------------------------------------------------------*/

respond_to_file(error):-!.
respond_to_file(Message):-
	find_performative(Message,P),
	action_required(P),!,
	do_requested_action(Message,ReplyMessages),
	send_messages(ReplyMessages),!.

send_messages([]):-!.
send_messages([ReplyMessage|RestMessages]):-
	output_file(ReplyMessage,File),
	tell(File),
	writemessage(ReplyMessage),
	told,!,
	send_messages(RestMessages).		

do_requested_action(Message,ReplyMessages):-
	find_language(Message,L),
	not speak(L),!,
	compose(sorry,Message,ReplyMessages).
do_requested_action(Message,ReplyMessages):-
	find_performative(Message,P),!,
	compose(P,Message,ReplyMessages).

/*--------------------------------------------------------------------
These are predicates to find words in messages
--------------------------------------------------------------------*/

find_performative(message(performative(P),_,_,_,_,_,_),P).
find_sender(message(_,sender(S),_,_,_,_,_),S).
find_receiver(message(_,_,receiver(R),_,_,_,_),R).
find_replyto(message(_,_,_,replyto(RT),_,_,_),RT).
find_replywith(message(_,_,_,_,replywith(RW),_,_),RW).
find_language(message(_,_,_,_,_,language(L),_),L).
find_content(message(_,_,_,_,_,_,content(C)),C).

/*--------------------------------------------------------------------
Messages in which the language is not spoken are replied by "sorry"
--------------------------------------------------------------------*/

compose(sorry,Message,[NewMessage]):-!,
	find_sender(Message,S),
	find_replywith(Message,RW),
	get_agent_id(A),
	get_message_id(ID),
	NewMessage=message(	performative(sorry),
				sender(A),
				receiver(S),
				replyto(RW),
				replywith(ID),
				language(empty),
				content(empty)).

/*--------------------------------------------------------------------
Messages 'ask-if' should verify the existence of content in the KB
i.e. a simple "true" or "fail" is answered back
--------------------------------------------------------------------*/

compose('ask-if',Message,[NewMessage]):-
	find_sender(Message,S),
	find_replywith(Message,RW),
	find_content(Message,C),
	find_language(Message,L),
	get_agent_id(A),
	get_message_id(ID),
	string_to_term(C,Call),
	verify(Call,Value),
	NewMessage=message(	performative(tell),
				sender(A),
				receiver(S),
				replyto(RW),
				replywith(ID),
				language(L),
				content(Value)).

/*--------------------------------------------------------------------
Messages 'ask-one' should find one answer to the query from the KB
i.e. an "<instantiation>" or "fail" is answered back
--------------------------------------------------------------------*/

compose('ask-one',Message,[NewMessage]):-
	find_sender(Message,S),
	find_replywith(Message,RW),
	find_content(Message,C),
	find_language(Message,L),
	get_agent_id(A),
	get_message_id(ID),
	string_to_term(C,Call),
	find_one(Call,Value),
	NewMessage=message(	performative(tell),
				sender(A),
				receiver(S),
				replyto(RW),
				replywith(ID),
				language(L),
				content(Value)).


/*--------------------------------------------------------------------
Messages 'ask-all' should find all answers to the query from the KB
i.e. a list of "<instantiation>" or "fail" is answered back
--------------------------------------------------------------------*/

compose('ask-all',Message,[NewMessage]):-
	find_sender(Message,S),
	find_replywith(Message,RW),
	find_content(Message,C),
	find_language(Message,L),
	get_agent_id(A),
	get_message_id(ID),
	string_to_term(C,Call),
	find_all(Call,Value),
	NewMessage=message(	performative(tell),
				sender(A),
				receiver(S),
				replyto(RW),
				replywith(ID),
				language(L),
				content(Value)).

/*--------------------------------------------------------------------
Messages 'advertise' should just return the content which is a
KQML message itself
--------------------------------------------------------------------*/
compose('advertise',Message,[Content]):-!,
	find_content(Message,Content).

compose('broadcast',Message,NewMessages):-
	find_content(Message,C),
	find_performative(C,P),		% find all these in C
	find_sender(C,S),		% which is the message
	find_replywith(C,RW),		% to be broadcasted
	find_content(C,CC),
	find_language(C,L),
	known_agents(AgentList),
	findall(NewMessage,
			(member(Agent,AgentList),
			get_message_id(ID),
			NewMessage=message(performative(P),
				sender(S),
				receiver(Agent),
				replyto(RW),
				replywith(ID),
				language(L),
				content(CC))),
		NewMessages),!.	

/*--------------------------------------------------------------------
Auxiliary predicate definitions used from above
--------------------------------------------------------------------*/

find_all(C,Value):-
	findall(C,C,L),
	(L=[] -> Value=fail ; Value=L).

find_one(Call,Call):-
	Call,!.
find_one(_,fail).

verify(Call,true):-
	Call,!.
verify(_,fail).

/*--------------------------------------------------------------------
writelist/1: takes a term (meesage/7) and writes it to the std output
--------------------------------------------------------------------*/

writemessage(message(	performative(P),
			sender(S),
			receiver(R),
			replyto(RT),
			replywith(RW),
			language(L),
			content(C))):-
%	nl,write('The answer to the Message is:'),nl,
	write('('),write(P),nl,
	tab(10),write(':'),write(sender),tab(1),write(S),nl,
	tab(10),write(':'),write(receiver),tab(1),write(R),nl,	
	tab(10),write(':'),write(in-reply-to),tab(1),write(RT),nl,	
	tab(10),write(':'),write(reply-with),tab(1),write(RW),nl,
	(L\=empty ->
		tab(10),write(':'),write(language),tab(1),write(L),nl;
		true),
	(C\=empty ->
		tab(10),write(':'),write(content),tab(1),
		write('"'),write(C),write('"'),nl;
		true),
	write(').'),nl,!.

/*--------------------------------------------------------------------
Performatives for which action is required
--------------------------------------------------------------------*/

action_required('ask-if').
action_required('ask-all').
action_required('ask-one').
action_required('advertise').
action_required('broadcast').

/*--------------------------------------------------------------------
Performatives for which action is NOT required
--------------------------------------------------------------------*/

response_only('tell').

no_response('sorry').


/*--------------------------------------------------------------------
The name of the output file should be 'Receiver/MessageID.txt'
--------------------------------------------------------------------*/

output_file(ReplyMessage,OFile):-
	find_receiver(ReplyMessage,Receiver),
	find_replywith(ReplyMessage,ID),
	string_to_atom(S1,Receiver),
	string_to_atom(S2,ID),
	string_to_atom(S3,'/'),
	string_to_atom(S4,'.txt'),
	string_concat(S1,S3,T1),
	string_concat(S2,S4,T2),
	string_concat(T1,T2,SF),
	string_to_atom(SF,OFile),!.

/*--------------------------------------------------------------------
  hack!! hack!! hack!! beacuse of ".." need to be in content
--------------------------------------------------------------------*/

string_to_term(S,T):-
	tell(tmp),
	write(S),write('.'),nl,
	told,
	see(tmp),
	read(T),
	seen,!.
